class DerivaUploadError (RuntimeError):
    pass


class DerivaUploadConfigurationError (ValueError):
    pass


class DerivaUploadCatalogCreateError (RuntimeError):
    pass


class DerivaUploadCatalogUpdateError (RuntimeError):
    pass
